import sqlalchemy as sql
import pandas as pd
import importlib.resources as resources

SCHEMA_VERSION = "0.2"
GENETMAP_PATH='genetmap.csv'
GENETMAP_DTYPE={'chromosome': str, 'position': int, 'cm': float}

@sql.event.listens_for(sql.engine.Engine, "connect")
def set_sqlite_pragma(dbapi_connection, connection_record):
    cursor = dbapi_connection.cursor()
    cursor.execute("PRAGMA foreign_keys=ON")
    cursor.close()

metadata = sql.MetaData()

meta = sql.Table(
    "meta",
    metadata,
    sql.Column("key", sql.String, nullable=False, primary_key=True),
    sql.Column("value", sql.String, nullable=False),
)

kit = sql.Table(
    "kit",
    metadata,
    sql.Column("id", sql.Integer, nullable=False, primary_key=True),
    sql.Column("kitid", sql.String, nullable=False, unique=True, sqlite_on_conflict_unique='IGNORE'),
    sql.Column("name", sql.String),
    sql.Column("email", sql.String),
    sql.Column("sex", sql.String),
)

genetmap = sql.Table(
    "genetmap",
    metadata,
    sql.Column("chromosome", sql.String, nullable=False, index=True),
    sql.Column("position", sql.Integer, nullable=False, index=True),
    sql.Column("cm", sql.Float, nullable=False),
    sql.UniqueConstraint("chromosome", "position"),
)

source = sql.Table(
    "source",
    metadata,
    sql.Column("kit", sql.Integer, sql.ForeignKey("kit.id"), nullable=False, primary_key=True,
               sqlite_on_conflict_primary_key="IGNORE"),
    sql.Column("match", sql.Integer), # batch number of most recent matches
    sql.Column("triangle", sql.Integer), # batch number of most recent triangles
    sql.Column("negative", sql.Integer), # negative triangles incorporate data from this batch
)

segment = sql.Table(
    "segment",
    metadata,
    sql.Column("id", sql.Integer, nullable=False, primary_key=True),
    sql.Column("chromosome", sql.String, nullable=False, index=True), # chromosome 23 is "X"
    sql.Column("start", sql.Integer, nullable=False),
    sql.Column("end", sql.Integer, nullable=False),
    sql.Column("length", sql.Float, index=True),
    sql.UniqueConstraint("chromosome", "start", "end", sqlite_on_conflict='IGNORE'),
)

# pairwise matches are included with both permutations of (kit1, kit2)
match = sql.Table(
    "match",
    metadata,
    sql.Column("segment", sql.Integer, sql.ForeignKey("segment.id"), nullable=False, index=True),
    sql.Column("kit1", sql.Integer, sql.ForeignKey("kit.id"), nullable=False, index=True),
    sql.Column("kit2", sql.Integer, sql.ForeignKey("kit.id"), nullable=False, index=True),
    sql.Column("batch", sql.Integer, nullable=False, index=True), # when it was added
    sql.UniqueConstraint("segment", "kit1", "kit2", sqlite_on_conflict='IGNORE'),
)

# triangles are included with all 6 permutations of (kit1, kit2, kit3)
triangle = sql.Table(
    "triangle",
    metadata,
    sql.Column("segment", sql.Integer, sql.ForeignKey("segment.id"), nullable=False, index=True),
    sql.Column("kit1", sql.Integer, sql.ForeignKey("kit.id"), nullable=False, index=True),
    sql.Column("kit2", sql.Integer, sql.ForeignKey("kit.id"), nullable=False, index=True),
    sql.Column("kit3", sql.Integer, sql.ForeignKey("kit.id"), nullable=False, index=True),
    sql.Column("batch", sql.Integer, nullable=False, index=True), # when it was added
    sql.UniqueConstraint("segment", "kit1", "kit2", "kit3", sqlite_on_conflict='IGNORE'),
)

# gives segments where (source, target1) and (source, target2) have overlapping matches
# used for finding negative triangles
overlap = sql.Table(
    "overlap",
    metadata,
    sql.Column("id", sql.Integer, nullable=False, primary_key=True),
    sql.Column("source", sql.Integer, sql.ForeignKey("source.kit"), nullable=False, index=True),
    sql.Column("target1", sql.Integer, sql.ForeignKey("kit.id"), nullable=False, index=True),
    sql.Column("target2", sql.Integer, sql.ForeignKey("kit.id"), nullable=False, index=True),
    sql.Column("segment", sql.Integer, sql.ForeignKey("segment.id"), nullable=False, index=True),
    sql.UniqueConstraint("source", "target1", "target2", "segment", sqlite_on_conflict='IGNORE'),
)

negative = sql.Table(
    "negative",
    metadata,
    sql.Column("overlap", sql.Integer, sql.ForeignKey("overlap.id", ondelete="CASCADE"),
               nullable=False, index=True),
    sql.Column("neg_segment", sql.Integer, sql.ForeignKey("segment.id", ondelete="CASCADE"),
               nullable=False, index=True),
    sql.UniqueConstraint("overlap", "neg_segment", sqlite_on_conflict='IGNORE'),
)

def initialize(engine):
    metadata.drop_all(engine)
    metadata.create_all(engine)

    kg_meta_init = [
        dict(key='schema_version', value=str(SCHEMA_VERSION)),
        dict(key='batch', value=str(0)),
    ]

    import kgenealogic as kg
    genetmap_df = pd.read_csv(resources.files(kg).joinpath(GENETMAP_PATH), dtype=GENETMAP_DTYPE)

    with engine.connect() as conn:
        conn.execute(meta.insert(), kg_meta_init)
        conn.execute(genetmap.insert(), genetmap_df.to_dict(orient="records"))
        conn.commit()



import sqlalchemy as sql
import pandas as pd

from kgenealogic.schema import *

def increment_batch(engine):
    """Increment batch number in meta table and return new value."""
    select_batch = (
        sql.select(meta.c.value)
        .where(meta.c.key=='batch')
    )
    update_batch = (
        sql.update(meta)
        .where(meta.c.key=='batch')
    )
    with engine.connect() as conn:
        batch = int(conn.execute(select_batch).scalar_one())
        batch += 1
        conn.execute(update_batch, [dict(value=str(batch))])
        conn.commit()
    return batch

def add_sources(engine, sources):
    """Add sources to source table, if not already present."""
    with engine.connect() as conn:
        conn.execute(source.insert(), [dict(kit=int(x)) for x in sources])
        conn.commit()

def set_segment_lengths(engine):
    """Compute cM lengths for segments using dnapainter genetic map."""
    seg_pos = (
        sql.select(
            segment.c["id", "chromosome", "start", "end"],
            sql.func.max(genetmap.c.position).filter(genetmap.c.position <= segment.c.start).label("start1"),
            sql.func.min(genetmap.c.position).filter(genetmap.c.position >= segment.c.start).label("start2"),
            sql.func.max(genetmap.c.position).filter(genetmap.c.position <= segment.c.end).label("end1"),
            sql.func.min(genetmap.c.position).filter(genetmap.c.position >= segment.c.end).label("end2"),
        )
        .join_from(segment, genetmap, segment.c.chromosome==genetmap.c.chromosome)
        .where(segment.c.length.is_(None))
        .group_by(segment.c["id", "chromosome"])
        .cte()
    )

    cm_s1, cm_s2, cm_e1, cm_e2 = genetmap.alias(), genetmap.alias(), genetmap.alias(), genetmap.alias()

    seg_length = (
        sql.select(
            seg_pos.c.id,
            ((cm_e1.c.cm + sql.func.coalesce(((seg_pos.c.end - cm_e1.c.position)/(cm_e2.c.position-cm_e1.c.position))*(cm_e2.c.cm-cm_e1.c.cm),0)) -
            (cm_s1.c.cm + sql.func.coalesce(((seg_pos.c.start - cm_s1.c.position)/(cm_s2.c.position-cm_s1.c.position))*(cm_s2.c.cm-cm_s1.c.cm),0))).label("length")
        )
        .join_from(seg_pos, cm_s1, sql.and_(
            cm_s1.c.chromosome==seg_pos.c.chromosome,
            cm_s1.c.position==sql.func.coalesce(seg_pos.c.start1, seg_pos.c.start2)
        ))
        .join(cm_s2, sql.and_(
            cm_s2.c.chromosome==seg_pos.c.chromosome,
            cm_s2.c.position==sql.func.coalesce(seg_pos.c.start2, seg_pos.c.start1)
        ))
        .join(cm_e1, sql.and_(
            cm_e1.c.chromosome==seg_pos.c.chromosome,
            cm_e1.c.position==sql.func.coalesce(seg_pos.c.end1, seg_pos.c.end2)
        ))
        .join(cm_e2, sql.and_(
            cm_e2.c.chromosome==seg_pos.c.chromosome,
            cm_e2.c.position==sql.func.coalesce(seg_pos.c.end2, seg_pos.c.end1)
        ))
        .subquery()
    )
        
    update_seg_length = (
        sql.update(segment)
        .where(segment.c.id==seg_length.c.id)
        .values(length=seg_length.c.length)
    )
    with engine.connect() as conn:
        conn.execute(update_seg_length)
        conn.commit()

def as_internal_kitid(engine, data, kitid_fields):
    """Replace external kitid with internal kit number, inserting into kit table as needed."""
    if len(data)==0:
        return data

    kit_ids = pd.concat([data[c] for c in kitid_fields]).drop_duplicates()
    insert_kit_id = kit.insert().values(kitid=sql.bindparam('kit'))
    select_kit_id = sql.select(kit.c["id", "kitid"])
    with engine.connect() as conn:
        conn.execute(insert_kit_id, [dict(kit=x) for x in kit_ids])
        conn.commit()
        kit_ids = pd.read_sql(select_kit_id, conn)

    kit_ids = kit_ids.set_index('kitid')['id']
    for c in kitid_fields:
        data = data.assign(**{c: data[c].map(kit_ids)})
    return data

def as_internal_segment(engine, data):
    """Replace chromosome/start/end triplet with internal segment id, inserting into segment table
    as needed."""
    if len(data)==0:
        return data.assign(segment=-1)

    seg_df = (
        data[["chromosome", "start", "end"]]
        .groupby(["chromosome", "start", "end"])
        .head(1)
    )
    select_segment_ids = sql.select(segment.c.id.label("segment"), segment.c["chromosome", "start", "end"])
    with engine.connect() as conn:
        conn.execute(segment.insert(), seg_df.to_dict(orient="records"))
        conn.commit()
        segment_ids = pd.read_sql(select_segment_ids, conn)

    set_segment_lengths(engine)

    data = data.merge(segment_ids, on=["chromosome", "start", "end"])
    return data

def update_kit_data(engine, kit_data):
    """Update name/email/sex fields of kit table"""
    insert_kit_data = (
        kit.update()
        .where(kit.c.id==sql.bindparam("kit"))
        .where(kit.c.sex.is_(None))
        .values(
            name=sql.bindparam("name"),
            email=sql.bindparam("email"),
            sex=sql.bindparam("sex"),
        )
    )
    with engine.connect() as conn:
        conn.execute(insert_kit_data, kit_data.to_dict(orient='records'))
        conn.commit()

def import_matches(engine, data):
    """Import pairwise matches to project database from dataframe.

    Args:
        engine: the sqlalchemy engine for the project database
        data: a pandas dataframe with string fields kit1, kit2, chromosome, start, end, name,
        email, sex. Every kit appearing in the kit1 field is assumed to be a "source" for the
        matches
    """
    data = as_internal_kitid(engine, data, ['kit1', 'kit2'])

    kit_data = (
        data[['kit2', 'name', 'email', 'sex']]
        .groupby(['kit2'])
        .head(1)
        .drop_duplicates()
        .rename(columns=dict(kit2='kit'))
    )
    update_kit_data(engine, kit_data)

    sources = data['kit1'].drop_duplicates()
    add_sources(engine, sources)

    data = as_internal_segment(engine, data)

    batch = increment_batch(engine)
    data['batch'] = batch
    data = data[['kit1', 'kit2', 'segment', 'batch']]
    data = pd.concat([data, data.rename(columns=dict(kit1='kit2', kit2='kit1'))],
                        ignore_index=True)

    updated_kits = (
        sql.select(match.c.kit1.distinct().label("kit"))
        .where(match.c.batch==batch)
        .subquery()
    )
    update_sources = (
        sql.update(source)
        .where(source.c.kit==updated_kits.c.kit)
        .values(match=batch)
    )

    with engine.connect() as conn:
        conn.execute(match.insert(), data.to_dict(orient="records"))
        conn.execute(update_sources)
        conn.commit()

def import_triangles(engine, data):
    """Import triangulations to project database from dataframe.

    Args:
        engine: the sqlalchemy engine for the project database
        data: a pandas dataframe with string fields kit1, kit2, kit3, chromosome, start, end, name,
        email. Every kit appearing in the kit1 field is assumed to be a "source" for the triangles
    """
    tri_df = as_internal_kitid(engine, data, ['kit1', 'kit2', 'kit3'])

    kit_data = (
        pd.concat([
            tri_df[['kit2', 'name2', 'email2']].rename(
                columns=dict(kit2='kit', name2='name', email2='email')
            ),
            tri_df[['kit3', 'name3', 'email3']].rename(
                columns=dict(kit3='kit', name3='name', email3='email')
            ),
        ], ignore_index=True)
        .dropna()
        .drop_duplicates()
    )
    kit_data['sex'] = None
    update_kit_data(engine, kit_data)

    sources = tri_df['kit1'].drop_duplicates()
    add_sources(engine, sources)

    tri_df = as_internal_segment(engine, tri_df)

    batch = increment_batch(engine)
    tri_df['batch'] = batch
    tri_df = tri_df[['kit1', 'kit2', 'kit3', 'segment', 'batch']]
    tri_df = pd.concat([
            tri_df,
            tri_df.rename(columns=dict(kit1='kit2', kit2='kit1')),
            tri_df.rename(columns=dict(kit1='kit3', kit3='kit1')),
            tri_df.rename(columns=dict(kit2='kit3', kit3='kit2')),
            tri_df.rename(columns=dict(kit1='kit2', kit2='kit3', kit3='kit1')),
            tri_df.rename(columns=dict(kit1='kit3', kit2='kit1', kit3='kit2')),
        ], ignore_index=True)

    updated_kits = (
        sql.select(triangle.c.kit1.distinct().label("kit"))
        .where(triangle.c.batch==batch)
        .subquery()
    )
    update_sources = (
        sql.update(source)
        .where(source.c.kit==updated_kits.c.kit)
        .values(triangle=batch)
    )
    with engine.connect() as conn:
        conn.execute(triangle.insert(), tri_df.to_dict(orient="records"))
        conn.execute(update_sources)
        conn.commit()

__all__ = [import_matches, import_triangles]



import numpy as np
import scipy as sp
import pandas as pd
import sqlalchemy as sql
from dataclasses import dataclass, field
import typing
import typer

from kgenealogic.schema import *

# factor for weighting pairwise matches in the graph; triangles have weight 1
PAIRWISE_FACTOR = 0.25

@dataclass
class Seed:
    kit: int
    floating: bool = field(default_factory=bool)
    negative: bool = field(default_factory=bool)

@dataclass
class SeedTree:
    """Data structure for organizing cluster seeds in tree"""
    ahnentafel: int
    values: list[Seed] = field(default_factory=list)
    branches: dict[str, typing.Any] = field(default_factory=dict)

def flatten(tree, seeds=None):
    """Flatten a SeedTree by putting internal kit ids into values list."""
    if seeds is None:
        seeds = {}
    for b in tree.branches.values():
        flatten(b, seeds)
    seeds.update({v.kit: tree.ahnentafel for v in tree.values})
    return seeds

def cluster_data(engine, config):
    """Cluster data according to the config file

    The behavior is documented in the CLI cluster command.

    Args:
        engine: the sqlalchemy engine for the project database
        config: a ClusterConfig object describing the clustering to perform

    Returns:
        a pandas dataframe classifying kits, as described in the CLI cluster command documentation
    """
    with engine.connect() as conn:
        kits = pd.read_sql(sql.select(kit.c['id', 'kitid']), conn)
    kits = kits.set_index('kitid').id

    seeds = set()
    exclude = set()
    for kitid_list, kit_set in ((config.seeds, seeds), (config.exclude, exclude)):
        for kitid in kitid_list:
            if kitid not in kits.index:
                typer.echo(f"Invalid kit id {kitid}", err=True)
                raise typer.Exit(code=1)
            else:
                kit_set.add(int(kits.loc[kitid]))

    if config.include:
        include = set()
        match_include_query = (
            sql.select(sql.distinct(match.c.kit2).label("kit"))
            .join_from(match, segment)
            .where(match.c.kit1==sql.bindparam("kit"))
            .where(segment.c.length > sql.bindparam("length"))
        )
        tri_include_query = (
            sql.select(sql.distinct(triangle.c.kit2).label("kit"))
            .join_from(triangle, segment)
            .where(triangle.c.kit1==sql.bindparam("kit"))
            .where(segment.c.length > sql.bindparam("length"))
        )
        with engine.connect() as conn:
            for k in config.include:
                kitid = k['id']
                if kitid not in kits.index:
                    typer.echo(f"Invliad kit id {kitid}", err=True)
                    raise typer.Exit(code=1)
                else:
                    include.add(int(kits.loc[kitid]))
                kitloc = int(kits.loc[kitid])
                if k.get('matches', None) is not None:
                    query_params = dict(kit=kitloc, length=k['matches'])
                    k_nbr = pd.read_sql(match_include_query, conn, params=query_params)
                    include.update([int(x) for x in k_nbr.kit])
                if k.get('triangles', None) is not None:
                    query_params = dict(kit=kitloc, length=k['triangles'])
                    k_nbr = pd.read_sql(tri_include_query, conn, params=query_params)
                    include.update([int(x) for x in k_nbr.kit])
        kits = kits[kits.isin(include|seeds)]

    kits = kits[~kits.isin(exclude)]

    graph_query = (
        sql.select(
            match.c.kit1,
            match.c.kit2,
            sql.func.sum(segment.c.length).label("weight"),
        )
        .join_from(match, segment, match.c.segment==segment.c.id)
        .where(segment.c.length >= config.min_length)
        .where(match.c.kit1 != match.c.kit2)
        .group_by(match.c['kit1', 'kit2'])
    )
    triangle_query = (
        sql.select(
            triangle.c.kit1,
            triangle.c.kit2,
            sql.func.sum(segment.c.length).label("weight"),
        )
        .join_from(triangle, segment, triangle.c.segment==segment.c.id)
        .where(segment.c.length >= config.min_length)
        .where(triangle.c.kit1 != triangle.c.kit2)
        .where(triangle.c.kit3.not_in(config.exclude))
        .group_by(triangle.c['kit1', 'kit2'])
    )
    trisource_query = (
        sql.select(source.c.kit)
        .where(source.c.triangle.is_not(None))
    )

    with engine.connect() as conn:
        graph = pd.read_sql(graph_query, conn)
        tri_graph = pd.read_sql(triangle_query, conn)
        trisource= set(int(x) for x in pd.read_sql(trisource_query, conn).kit)

    graph['weight'] = PAIRWISE_FACTOR*graph['weight'].astype(float)
    graph = (
        graph
        .set_index(['kit1', 'kit2'])
        .add(tri_graph.set_index(['kit1', 'kit2']), fill_value=0)
        .reset_index()
    )

    autox_query = (
        sql.select(sql.distinct(match.c.kit2).label('kit'))
        .join_from(match, segment, match.c.segment==segment.c.id)
        .where(segment.c.chromosome=='X')
        .where(segment.c.length >= config.min_length)
        .where(match.c.kit1==sql.bindparam('seed'))
    )
    def build_seed_tree(raw, ahnentafel):
        parsed = SeedTree(ahnentafel=ahnentafel)
        autox = set()
        for k in raw.get('kits', []):
            kitlocal = int(kits.loc[k['id']])
            floating = k.get('float')
            if floating is None:
                floating = kitlocal not in trisource
            if k.get('autox'):
                with engine.connect() as conn:
                    k_autox = pd.read_sql(autox_query, conn, params=dict(seed=kitlocal))
                k_autox = set(int(x) for x in k_autox.kit)
                k_autox.difference_update(exclude)
                k_autox.difference_update(seeds)
                seeds.update(k_autox)
                autox.update(k_autox)
            negative = k.get('negative', False)
            parsed.values.append(Seed(kit=kitlocal, floating=floating, negative=negative))

        for a, short, long in ((1, 'M', 'maternal'), (0, 'P', 'paternal')):
            if long in raw:
                parsed.branches[short] = build_seed_tree(raw[long],a+(2*ahnentafel))

        if autox:
            if 'M' not in parsed.branches:
                parsed.branches['M'] = SeedTree(values=[],ahnentafel=1+(2*ahnentafel))
            maternal_values = parsed.branches['M'].values
            for x in autox:
                maternal_values.append(Seed(kit=x, floating=True))
        return parsed
    tree = build_seed_tree(config.tree, 1)

    seed_labels = flatten(tree)

    source_neg = lambda s_kit: get_negative(engine, s_kit, config.min_length)
    clusters = recursive_cluster(kits, tree, graph, source_neg)
    
    clusters['seed'] = clusters.kit.map(seed_labels).astype('Int64')
    kitids = kits.reset_index().set_index('id').kitid
    clusters['kit'] = clusters.kit.map(kitids)

    fixed_cols = ['kit', 'ahnentafel', 'seed']
    clusters = clusters[fixed_cols + [c for c in clusters.columns if c not in fixed_cols]]

    return clusters

def recursive_cluster(kits, tree, graph, source_neg):
    """Recursively partition a list of kits using a given tree structure.

    Args:
        kits: pandas Series of internal kit ids to be be included in the segmentation
        tree: a SeedTree describing the structure of the segmentation and the seeds to be used
        graph: the a pandas DataFrame describing the graph of weighted relationships to be used to
            construct the segmentation, in sparse COO format. Columns 'kit1' and 'kit2' give internal
            kit numbers, and column 'weight' gives the weight of the edge between them. Symmetric,
            so 'kit1' and 'kit2' appear in both orders, if there is an edge between them.
        source_neg: a function mapping internal kit numbers to negative triangles in the same
            format as `graph`

    Returns:
        pandas DataFrame in the same format as the result of cluster_data, but with internal kit
        numbers instead of external kit id strings
    """
    nonfloat = []
    tri_graph = graph
    for s_kit in tree.values:
        if s_kit.negative:
            tri_graph = tri_graph.set_index(['kit1', 'kit2']).add(
                source_neg(s_kit.kit).set_index(['kit1', 'kit2']), fill_value=0
            ).reset_index()
        if not s_kit.floating:
            nonfloat.append(s_kit.kit)
        
    depth = int(np.log2(tree.ahnentafel))
    result = pd.DataFrame(dict(kit=kits))

    kits = kits[~kits.isin(nonfloat)]
    graph = graph[graph.kit1.isin(kits)&graph.kit2.isin(kits)]
    tri_graph = tri_graph[tri_graph.kit1.isin(kits)&tri_graph.kit2.isin(kits)]
    if tree.branches and (len(result)>0) and (len(tri_graph) > 0):
        flat_seeds = []
        for label, tree_branch in tree.branches.items():
            branch_seeds = flatten(tree_branch).keys()
            branch_seeds = pd.DataFrame(dict(kit=branch_seeds))
            branch_seeds['label'] = label
            flat_seeds.append(branch_seeds)
        flat_seeds = pd.concat(flat_seeds, ignore_index=True)

        clusters = get_clusters(tri_graph, flat_seeds).set_index('kit')
        result['label'] = (
            result.kit.map(clusters.label)
            .fillna('')
            .astype(str)
        )
        result['confidence'] = result.kit.map(clusters.confidence)
        result['ahnentafel'] = np.select(
            [result.label=='P', result.label=='M'],
            [2*tree.ahnentafel, 1+(2*tree.ahnentafel)], tree.ahnentafel
        )

        result_branches = []
        for label, kit_branch in result.groupby('label'):
            if label in tree.branches:
                tree_branch = tree.branches[label]
                branch = recursive_cluster(kit_branch.kit, tree_branch, graph, source_neg)
                branch[f"label{depth}"] = branch.kit.map(kit_branch.set_index('kit').label)
                branch[f"confidence{depth}"] = branch.kit.map(kit_branch.set_index('kit').confidence)
            else:
                branch = kit_branch.rename(columns=dict(
                    label=f"label{depth}",
                    confidence=f"confidence{depth}",
                ))
            result_branches.append(branch)
        result = pd.concat(result_branches, ignore_index=True)
    else:
        flat_ahnentafel = flatten(tree)
        result['ahnentafel'] = result.kit.map(flat_ahnentafel).fillna(tree.ahnentafel)

    return result

def get_clusters(graph, seeds, max_rounds=None, fix_seeds=True):
    """Greedily segment each component of graph to respect the labels of seeds."""
    graph = graph.copy()
    labels = (
        graph
        .assign(weight=graph.weight.abs())
        .groupby('kit1')
        .weight
        .sum()
        .fillna(0)
        .reset_index()
        .rename(columns=dict(kit1='kit'))
    )
    labels['label'] = labels.kit.map(seeds.set_index('kit').label)
    labels = labels[labels.weight>0].set_index('kit').sort_index()
    
    max_rounds = max_rounds or 2*len(labels)
    i=0
    while True:
        graph['label'] = graph.kit2.map(labels.label)
        
        # TODO would be nice if seg_sum took confidence-weighted sum, but this is circular. does it converge?
        seg_sum = graph.groupby(['kit1', 'label']).weight.sum().unstack(level=-1).fillna(0)
        for l in ['P', 'M']:
            if l not in seg_sum:
                seg_sum[l] = 0
        labels['paternal'] = seg_sum.P-seg_sum.M
        labels['confidence'] = labels['paternal'].abs()/labels['weight']
        available = np.any([
            labels.label.isnull(),
            (labels.label=='P')&(labels.paternal<0),
            (labels.label=='M')&(labels.paternal>0),
        ], axis=0)
        available = available & (labels.confidence > 0)
        if fix_seeds:
            available = available & (~labels.index.isin(seeds.kit))
        i+=1
        
        if (not np.any(available)) or (i > max_rounds):
            break

        next_kit = labels[available].iloc[labels.confidence[available].argmax()]
        next_label = 'P' if next_kit.paternal>0 else 'M'
        labels.loc[next_kit.name, 'label'] = next_label

    labels['label'] = labels.label.fillna('').astype(str)
    labels = labels.reset_index()[['kit', 'label', 'confidence']]
    isolated_seeds = seeds[~seeds.kit.isin(labels.kit)]
    return pd.concat([labels, isolated_seeds], ignore_index=True)

def get_negative(engine, s_kit, min_length):
    """Get all negative triangles for a kit s_kit of length at least min_length."""
    build_negative(engine, s_kit)

    neg_tri = (
        sql.select(
            overlap.c.target1.label("kit1"),
            overlap.c.target2.label("kit2"),
            sql.func.sum(-segment.c.length).label("weight"),
        )
        .join_from(negative, overlap)
        .join(segment, negative.c.neg_segment==segment.c.id)
        .where(overlap.c.source==s_kit)
        .where(segment.c.length >= min_length)
        .group_by(overlap.c['target1', 'target2'])
    )

    with engine.connect() as conn:
        result = pd.read_sql(neg_tri, conn)
    return result

def build_negative(engine, s_kit):
    """Check if negative triangles are cached for source s_kit, and build if necessary."""
    with engine.connect() as conn:
        status = conn.execute(sql.select(source).where(source.c.kit==s_kit)).mappings().one_or_none()
    if (not status) or (not status['match']) or (not status['triangle']):
        # missing triangles or matches
        return False

    batch = max(status['match'], status['triangle'])
    neg_batch = status['negative'] or 0
    if neg_batch >= batch:
        # up to date
        return True

    update_source = sql.update(source).where(source.c.kit==s_kit).values(negative=batch)
    from kgenealogic.data import as_internal_segment

    valid_neg_targets = (
        sql.select(
            triangle.c.kit2.label("target"),
        )
        .where(triangle.c.kit1==s_kit)
        .distinct()
        .cte()
    )

    valid_matches = (
        sql.select(
            valid_neg_targets.c.target,
            segment.c.chromosome,
            segment.c.start,
            segment.c.end,
        )
        .join_from(valid_neg_targets, match, sql.and_(
            match.c.kit1==s_kit, match.c.kit2==valid_neg_targets.c.target
        ))
        .join(segment, match.c.segment==segment.c.id)
        .cte()
    )

    m1, m2 = valid_matches.alias(), valid_matches.alias()

    select_overlap = (
        sql.select(
            m1.c.target.label("target1"),
            m2.c.target.label("target2"),
            m1.c.chromosome,
            sql.func.max(m1.c.start, m2.c.start).label("start"),
            sql.func.min(m1.c.end, m2.c.end).label("end"),
        )
        .join_from(m1, m2, m1.c.chromosome==m2.c.chromosome)
        .where(m1.c.target!=m2.c.target)
        .where(m1.c.start < m2.c.end)
        .where(m2.c.start < m1.c.end)
    )

    with engine.connect() as conn:
        conn.execute(sql.delete(overlap).where(overlap.c.source==s_kit))
        match_overlap = pd.read_sql(select_overlap, conn)
    if len(match_overlap)==0:
        return True
    match_overlap = as_internal_segment(engine, match_overlap)

    match_overlap = match_overlap[["target1", "target2", "segment"]]
    match_overlap["source"] = s_kit
    with engine.connect() as conn:
        conn.execute(overlap.insert(), match_overlap.to_dict(orient="records"))
        conn.commit()
        
    seg_over, seg_tri = segment.alias(), segment.alias()
    select_neg_overlap = (
        sql.select(
            overlap.c.id.label("overlap"),
            seg_over.c.chromosome,
            seg_over.c.start,
            seg_over.c.end,
            seg_tri.c.start.label("tri_start"),
            seg_tri.c.end.label("tri_end"),
        )
        .join_from(overlap, seg_over, overlap.c.segment==seg_over.c.id)
        .join(triangle, sql.and_(
            overlap.c.source==triangle.c.kit1,
            overlap.c.target1==triangle.c.kit2,
            overlap.c.target2==triangle.c.kit3,
        ), isouter=True)
        .join(seg_tri, triangle.c.segment==seg_tri.c.id, isouter=True)
        .where(sql.or_(
            seg_tri.c.id.is_(None),
            sql.and_(
                seg_tri.c.chromosome==seg_over.c.chromosome,
                seg_tri.c.start <= seg_over.c.end,
                seg_tri.c.end >= seg_over.c.start,
            ),
        ))
        .where(overlap.c.source==s_kit)
    )
    with engine.connect() as conn:
        neg_overlap = pd.read_sql(select_neg_overlap, conn)

    if len(neg_overlap)==0:
        return True
    neg_base = (
        neg_overlap[neg_overlap.tri_start.isnull()]
        .drop(columns=["tri_start", "tri_end"])
        .drop_duplicates()
    )
    neg_remain = []
    neg_group_cols = ["overlap", "chromosome"]
    for _, tri_data in neg_overlap.dropna().groupby(neg_group_cols):
        id_vals = tri_data[neg_group_cols].iloc[0].to_dict()
        start, end = tri_data[['start', 'end']].iloc[0]
        # negative triangulations are the raw triangulation (start to end),
        # excluding all the positive parts (tri_start to tri_end)
        for _, row in tri_data.sort_values('tri_start').iterrows():
            if row.tri_start > start:
                neg_remain.append(dict(
                    start=start,
                    end=int(row.tri_start),
                    **id_vals))
            start = int(row.tri_end)
        if end > start:
             neg_remain.append(dict(
                start=start,
                end=end,
                **id_vals))       
    neg_tri = pd.concat([neg_base, pd.DataFrame(neg_remain)], ignore_index=True).drop_duplicates()
    neg_tri = as_internal_segment(engine, neg_tri).drop_duplicates()
    neg_tri = neg_tri[["overlap", "segment"]].rename(columns=dict(segment="neg_segment"))
    if len(neg_tri)==0:
        return True
    with engine.connect() as conn:
        conn.execute(negative.insert(), neg_tri.to_dict(orient="records"))
        conn.execute(update_source)
        conn.commit()

    return True



import pandas as pd
import yaml
from collections import defaultdict
from dataclasses import dataclass, field
import typing
import typer

DEFAULT_MIN_LENGTH = 7.

GED_MATCH_COLS = {
    'PrimaryKit': 'kit1',
    'MatchedKit': 'kit2',
    'chr': 'chromosome',
    'B37Start': 'start',
    'B37End': 'end',
    'Segment cM': 'length',
    'MatchedName': 'name',
    'Matched Sex': 'sex',
    'MatchedEmail': 'email',
}
def is_ged_matches(path):
    """Check if file at path is a valid GEDmatch pairwise matches file."""
    try:
        csv = pd.read_csv(path, dtype=str)
    except:
        return False
    return set(GED_MATCH_COLS.keys()).issubset(csv.columns)

def read_ged_matches(path):
    """Load GEDmatch pairwise matches file into pandas dataframe."""
    dtype = defaultdict(lambda: str)
    dtype['B37Start'] = int
    dtype['B37End'] = int
    matches = pd.read_csv(path, dtype=dtype)
    return matches[GED_MATCH_COLS.keys()].rename(columns=GED_MATCH_COLS)

GED_TRIANGLE_COLS = {
    'Kit1 Number': 'kit2',
    'Kit1 Name': 'name2',
    'Kit1 Email': 'email2',
    'Kit2 Number': 'kit3',
    'Kit2 Name': 'name3',
    'Kit2 Email': 'email3',
    'Chr': 'chromosome',
    'B37 Start': 'start',
    'B37 End': 'end',
    'cM': 'length',
}
def is_ged_triangles(path):
    """Check if file at path is a valid GEDmatch triangulation file."""
    try:
        csv = pd.read_csv(path, dtype=str)
    except:
        return False
    return set(GED_TRIANGLE_COLS.keys()).issubset(csv.columns)

def read_ged_triangles(path, primary_kit):
    """Load GEDmatch triangulation file into pandas dataframe.

    Args:
        path: path to the file
        primary_kit: the kit from which the triangulations were generated

    Returns:
        A pandas dataframe in the format expected by data.import_triangles.
        Specifically, primary_kit will appear as kit1 in the dataframe
    """
    dtype = defaultdict(lambda: str)
    dtype['B37 Start'] = int
    dtype['B37 End'] = int
    tri = pd.read_csv(path, dtype=dtype)
    tri = tri[GED_TRIANGLE_COLS.keys()].rename(columns=GED_TRIANGLE_COLS)
    tri['kit1'] = primary_kit
    return tri

@dataclass
class ClusterConfig:
    """Data structure for storing parsed cluster configuration"""
    include: list[dict[str, typing.Any]]
    exclude: list[str]
    min_length: float
    tree: dict[str, typing.Any]
    seeds: set[str]

def expect_keys(d, l, error):
    bad_keys = set(d.keys())
    bad_keys.difference_update(l)
    if bad_keys:
        invalid = bad_keys.pop()
        typer.echo(error.format(invalid), err=True)
        raise typer.Exit(code=1)

def parse_config_tree(tree, seeds):
    """Recursively expand and extract all seeds from YAML config tree

    Seeds that appear in the config file as strings are expanded to dictionaries with default
    values for the keys and reinserted into the tree

    Args:
        tree: the dictionary representing the current node of the tree
        seeds: the list of all encountered seeds
    """
    expect_keys(tree, ['kits', 'maternal', 'paternal'], 
                "invalid YAML configuration format: invalid tree key {}")

    expanded_kits = []
    for k in tree.get('kits', []):
        if not k:
            typer.echo(f"invalid (false/missing) tree kits entry", err=True)
            raise typer.Exit(code=1)
        elif type(k)!=dict:
            expanded_kits.append(dict(id=str(k), autox=False, float=None, negative=False))
        else:
            expect_keys(k, ['id', 'autox', 'float', 'negative'],
                        "invalid YAML configuration format: invalid tree kits key {}")
            expanded_kits.append(k)
    tree['kits']=expanded_kits
    seeds.extend(expanded_kits)

    for branch in ('maternal', 'paternal'):
        if branch in tree:
            value = tree[branch]
            if value:
                if type(value)!=dict:
                    typer.echo(f"invalid {branch} branch: {value}", err=True)
                    raise typer.Exit(code=1)
                parse_config_tree(tree[branch], seeds)
            else:
                tree[branch] = {}

def read_cluster_config(path):
    """Parse a clustering config file in YAML format.

    The schema is given above in code, and is documented in the cluster CLI command.

    Args:
        path: path to the cluster config file

    Returns:
        A ClusterConfig object representing the contents of the config file
    """
    with open(path) as f:
        parsed = yaml.safe_load(f)

    if 'min_length' in parsed:
        min_length = parsed['min_length']
        if min_length != 0 and not min_length:
            min_length = DEFAULT_MIN_LENGTH
        else:
            min_length = float(min_length)
    else:
        min_length = DEFAULT_MIN_LENGTH

    tree = dict(parsed.get('tree') or {})
    exclude = list(parsed.get('exclude') or [])

    seeds = []
    parse_config_tree(tree, seeds)
    
    include = []
    include_raw = parsed.get('include', []) or []
    if type(include_raw)!=list:
        typer.echo(f"invalid include value: expected list", err=True)
        raise typer.Exit(code=1)
    for k in include_raw:
        if not k:
            typer.echo(f"invalid (false/missing) include list entry", err=True)
            raise typer.Exit(code=1)
        elif type(k)!=dict:
            include.append(dict(id=str(k), matches=None, triangles=None))
        else:
            expect_keys(k, ['id', 'matches', 'triangles'],
                        "invalid YAML configuration format: invalid include item key {}")
            include.append(k)

    unique_seeds = set()
    for s in seeds:
        if s['id'] in unique_seeds:
            typer.echo("duplicated seed {}".format(s['id']), err=True)
            raise typer.Exit(code=1)
        else:
            unique_seeds.add(s['id'])

    exclude_seeds = unique_seeds.intersection(exclude)
    if exclude_seeds:
        k = exclude_seeds.pop()
        typer.echo(f"excluded kit {k} is listed as seed", err=True)
        raise typer.Exit(code=1)

    result = ClusterConfig(
        include=include,
        exclude=exclude,
        min_length=min_length,
        tree=tree,
        seeds=unique_seeds,
    )
    return result

def write_clusters(clusters, path):
    """Write out the results of clustering to a CSV file."""
    clusters.to_csv(path, index=False)
